########### Standard ###########
import os
import sys

########### Local ###########
from orpytal.common import units
from orpytal import Orbit

########### External ###########
import untangle

##########################################################
# Values
##########################################################

AU = 149597870.7 * units.km
G = 6.67408e-11 / (1000 ** 3) * units.km ** 3 / (units.kg * units.s ** 2)

# km^3 / kg s^2


##########################################################
# Planets
##########################################################

mu_units = (units.km ** 3 / units.s ** 2)


class CentralBody(object):
    def __init__(self,
                 name,
                 radius,
                 mu,
                 a=None,
                 period=None,
                 e=None,
                 i=None,
                 parent=None,
                 id=None,
                 notes=None, **kwargs):
        self.name = name

        self.radius = radius * units.km if not isinstance(radius, units.Quantity) else radius.to(units.km)

        self.mu = mu * mu_units if not isinstance(mu, units.Quantity) else mu.to(mu_units)

        if a:
            self.a = a * units.km if not isinstance(a, units.Quantity) else a.to(units.km)

        if period:
            self.period = period * units.s if not isinstance(period, units.Quantity) else period.to(units.s)

        if e is not None:
            self.e = e * units.dimensionless if not isinstance(e, units.Quantity) else e.to(units.dimensionless)

        if i is not None:
            self.i = i * units.radians if not isinstance(i, units.Quantity) else i.to(units.radians)

        self.parent = parent

        if self.parent and isinstance(self.parent, CentralBody):
            self.orbit = Orbit(self.parent, '{}'.format(self.name),
                               a=self.a,
                               period=self.period,
                               e=self.e,
                               i=self.i,
                               **kwargs)
        if id:
            self.id = id

        if notes:
            self.notes = notes

    @classmethod
    def from_xml(cls, xml):
        id = int(xml.id.cdata)
        e = float(xml.ecc.cdata) * units.dimensionless
        mu = float(xml.gm.cdata) * (units.km ** 3 / units.s ** 2)
        a = float(xml.circ_r.cdata) * units.km
        i = float(xml.inc.cdata) * units.degree
        radius = float(xml.radius.cdata) * units.km
        name = xml.name.cdata
        notes = xml.notes.cdata
        parent = xml.parent.cdata

        return CentralBody(
            name,
            radius,
            mu,
            id=id,
            e=e,
            i=i,
            a=a,
            notes=notes,
            parent=None if parent == 'N/A' else parent
        )

    @classmethod
    def from_dict(cls, dict):
        return cls(name=dict['name'],
                   radius=dict['radius'],
                   mu=dict['mu'],
                   a=None if 'iad' not in dict else dict['a'],
                   period=None if 'period' not in dict else dict['period'],
                   e=None if 'e' not in dict else dict['e'],
                   i=None if 'i' not in dict else dict['i'],
                   parent=None if 'parent' not in dict else dict['parent'],
                   id=None if 'id' not in dict else dict['id'],
                   notes=None if 'notes' not in dict else dict['notes'],
                   )

    @classmethod
    def from_pickle(cls, filename):
        return cls.from_dict(utils.load_pickle(filename))

    def pickle(self, filename):
        utils.pickle_dict(filename, self.to_pickle_dict())

    def to_pickle_dict(self):
        dict = {
            'name': self.name,
            'radius': self.radius,
            'mu': self.mu
        }

        if self.a:
            dict['a'] = self.a
        if self.period:
            dict['period'] = self.period
        if self.e:
            dict['e'] = self.e
        if self.i:
            dict['i'] = self.i
        if self.parent:
            dict['parent'] = self.parent
        if self.id:
            dict['id'] = self.id
        if self.notes:
            dict['notes'] = self.notes

        return dict


class BodiesDict(object):
    def __init__(self, xml_file):

        self.__dict__ = {}

        all_bodies_xml = untangle.parse(xml_file)
        for body_xml in all_bodies_xml.body_data.body:
            body = CentralBody.from_xml(body_xml)
            self[body.name.lower()] = body

        for key, value in self.__dict__.items():
            if value.parent:
                value.parent = self[value.parent]

    def __getitem__(self, item):
        return self.__dict__[item.lower()]

    def __setitem__(self, key, value):
        self.__dict__[key.lower()] = value

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

BODIES = BodiesDict(os.path.join(os.path.dirname(__file__), 'data/body_data.xml'))

# Add periods
BODIES['MERCURY'].period = 7600531.965 * units.seconds
BODIES['VENUS'].period = 19414144.04 * units.seconds
BODIES['MOON'].period = 236059 * units.seconds
BODIES['EARTH'].period = 31574658.69 * units.seconds
BODIES['PHOBOS'].period = 27552.96 * units.seconds
BODIES['DEIMOS'].period = 109071.36 * units.seconds
BODIES['MARS'].period = 59359794.38 * units.seconds
BODIES['IO'].period = 152853.5047 * units.seconds
BODIES['EUROPA'].period = 306822.0384 * units.seconds
BODIES['GANYMEDE'].period = 618153.3757 * units.seconds
BODIES['JUPITER'].period = 374869015.7 * units.seconds
BODIES['TITAN'].period = 1377648 * units.seconds
BODIES['SATURN'].period = 929260480.9 * units.seconds
BODIES['TITANIA'].period = 752218.6176 * units.seconds
BODIES['URANUS'].period = 2656887761 * units.seconds
BODIES['TRITON'].period = .5877 * 3600 * units.seconds
BODIES['NEPTUNE'].period = 5193323332 * units.seconds
BODIES['CHARON'].period = 551856.7066 * units.seconds
BODIES['PLUTO'].period = 7753066072 * units.seconds

BODIES['CERES'] = CentralBody(
    name='Ceres',
    radius=476.20,
    mu=63.2,
    a=413690604,
    period=145238090.6,
    e=0.0757972598,
    i=10.593981 * units.degrees,
    parent=BODIES['SUN']
)

BODIES['CALLISTO'] = CentralBody(
    name='Callisto',
    radius=2410.30,
    mu=7179.289,
    a=1882700,
    period=1441931.19,
    e=0.0074,
    i=0.192 * units.degrees,
    parent=BODIES['JUPITER']
)

BODIES_532 = {}

################## Sun ##################

BODIES_532['SUN'] = CentralBody(
    name='Sun',
    radius=695990.00,
    mu=132712200000.00
)

################## Mercury ##################

BODIES_532['MERCURY'] = CentralBody(
    name='Mercury',
    radius=2439.70,
    mu=22032.09,
    a=57909043,
    period=7600531.965,
    e=0.205645,
    i=7.004161 * units.degrees

)

################## Venus ##################

BODIES_532['VENUS'] = CentralBody(
    name='Venus',
    radius=6051.80,
    mu=324858.592079,
    a=108208877,
    period=19414144.04,
    e=0.006766272,
    i=3.394656 * units.degrees
)
################## Earth ##################


BODIES_532['EARTH'] = CentralBody(
    name='Earth',
    radius=6378.137,
    mu=398600.4418,
    a=149649952,
    period=31574658.69,
    e=0.01621148,
    i=0.0008611136 * units.degrees
)

BODIES_532['MOON'] = CentralBody(
    name='Moon',
    radius=1737.40,
    mu=4902.801076,
    a=384400,
    period=2360592,
    e=0.0543866,
    i=5.16000 * units.degrees,
    parent=BODIES_532['EARTH']
)

################## Mars ##################

BODIES_532['MARS'] = CentralBody(
    name='Mars',
    radius=3396.19,
    mu=42828.3719012840,
    a=227953016,
    period=59359794.38,
    e=0.0933555537,
    i=1.848489
)

BODIES_532['PHOBOS'] = CentralBody(
    name='Phobos',
    radius=11.10,
    mu=0.0007112,
    a=9376,
    period=27552.96,
    e=0.01510000,
    i=1.075000,
    parent=BODIES_532['MARS']
)

BODIES_532['DEIMOS'] = CentralBody(
    name='Deimos',
    radius=6.20,
    mu=0.0000985,
    a=23458,
    period=109071.36,
    e=0.0002,
    i=1.78800,
    parent=BODIES_532['MARS']
)

################## Jupiter ##################

BODIES_532['JUPITER'] = CentralBody(
    name='Jupiter',
    radius=71492.00,
    mu=126686535,
    a=779067093,
    period=374869015.7,
    e=0.0490681913,
    i=1.303603 * units.degrees
)

BODIES_532['IO'] = CentralBody(
    name='Io',
    radius=1821.60,
    mu=5959.916,
    a=421800,
    period=152853.5047,
    e=0.0041,
    i=0.036 * units.degrees,
    parent=BODIES_532['JUPITER']
)

BODIES_532['EUROPA'] = CentralBody(
    name='Europa',
    radius=1560.80,
    mu=3202.739,
    a=671100,
    period=306822.0384,
    e=0.0094,
    i=0.466 * units.degrees,
    parent=BODIES_532['JUPITER']
)

BODIES_532['GANYMEDE'] = CentralBody(
    name='Ganymede',
    radius=2631.20,
    mu=9887.834,
    a=1070400,
    period=618153.3757,
    e=0.0013,
    i=0.177 * units.degrees,
    parent=BODIES_532['JUPITER']
)

BODIES_532['CALLISTO'] = CentralBody(
    name='Callisto',
    radius=2410.30,
    mu=7179.289,
    a=1882700,
    period=1441931.19,
    e=0.0074,
    i=0.192 * units.degrees,
    parent=BODIES_532['JUPITER']
)

################## Saturn ##################

BODIES_532['SATURN'] = CentralBody(
    name='Saturn',
    radius=60268.00,
    mu=37931284,
    a=1426647631,
    period=929260480.9,
    e=0.05483734,
    i=2.488476 * units.degrees
)

BODIES_532['TITAN'] = CentralBody(
    name='Titan',
    radius=2574.730,
    mu=8978.1382,
    a=1221865,
    period=1377648,
    e=0.0288,
    i=0.312 * units.degrees,
    parent=BODIES_532['SATURN']
)

################## Uranus ##################

BODIES_532['URANUS'] = CentralBody(
    name='Uranus',
    radius=25559.00,
    mu=5793965.66393928,
    a=2873682806,
    period=2656887761,
    e=0.046796657,
    i=0.77242 * units.degrees
)

BODIES_532['TITANIA'] = CentralBody(
    name='Titania',
    radius=788.90,
    mu=228.2,
    a=436300,
    period=752218.6176,
    e=0.0011,
    i=0.079 * units.degrees,
    parent=BODIES_532['URANUS']
)

################## Neptune ##################


BODIES_532['NEPTUNE'] = CentralBody(
    name='Neptune',
    radius=24764,
    mu=6835110.00,
    a=4492499814,
    period=5193323332,
    e=0.010016637,
    i=1.767227 * units.degrees
)

BODIES_532['TRITON'] = CentralBody(
    name='Triton',
    radius=1350.00,
    mu=1427.598,
    a=354759,
    period=.5877 * 3600,
    e=0.000016,
    i=156.86500 * units.degrees,
    parent=BODIES_532['NEPTUNE']
)

################## Pluto ##################


BODIES_532['PLUTO'] = CentralBody(
    name='Pluto',
    radius=1195.00,
    mu=873.7674473980320,
    a=5868125070,
    period=7753066072,
    e=0.247576314,
    i=17.325236 * units.degrees
)

BODIES_532['CHARON'] = CentralBody(
    name='Charon',
    radius=603.60,
    mu=102.3,
    a=17536,
    period=551856.7066,
    e=0.0022,
    i=0.001 * units.degrees,
    parent=BODIES_532['PLUTO']
)

################## Ceres ##################

BODIES_532['CERES'] = CentralBody(
    name='Ceres',
    radius=476.20,
    mu=63.2,
    a=413690604,
    period=145238090.6,
    e=0.0757972598,
    i=10.593981 * units.degrees

)

for k, b in BODIES.items():
    setattr(sys.modules[__name__], b.name.lower(), b)



########### Standard ###########
import math

########### Local ###########
from orpytal.common.utils import conics_utils

########### External ###########
import numpy as np
import scipy as sp


class KeplarianState(object):

    def __init__(self, orbit, name=''):
        self.orbit = orbit
        self.name = name

        self._ascending = None

        self._r = None
        self._v = None

        self._r_rth = None
        self._r_eph = None
        self._r_xyz = None

        self._v_rth = None
        self._v_eph = None
        self._v_xyz = None

        self._ta = None
        self._theta = None
        self._E = None
        self._M = None

        self._fpa = None
        self._ttp = None

        self._dcm_er = None
        self._dcm_rv = None
        self._dcm_ri = None

    def set_vars(self):
        self.r = None
        self.v = None

        self.r_rth = None
        self.r_eph = None
        self.r_xyz = None

        self.v_rth = None
        self.v_eph = None
        self.v_xyz = None

        self.ta = None
        self.theta = None
        self.E = None
        self.M = None
        self.fpa = None
        self.ttp = None

        self.dcm_er = None
        self.dcm_rv = None
        self.dcm_ri = None

        self.set_orbit()

    def set_orbit(self):
        self.orbit.p_from_state(self)
        self.orbit.e_from_state(self)
        self.orbit.se_from_state(self)
        self.orbit.h_from_state(self)
        self.orbit.h_xyz_from_state(self)

    @property
    def E(self):
        return self._E

    @E.setter
    def E(self, E=None):
        if self._E is not None or self.orbit.is_circular():
            return

        if E is not None:
            self._E = E
        elif self._M is not None:
            self._E = self._iteratively_find_E()
        elif self._ta is not None and self._r is not None and self.orbit.a is not None and self.orbit.e is not None:
            if self._ta == 0:
                self._E = 0
            elif self._ta == np.pi:
                self._E = np.pi
            else:
                self._E = self._get_ascending_sign* np.arccos(
                    (self.orbit.a - self._r) / (self.orbit.a * self.orbit.e))
        elif self.orbit.a is not None and self._r is not None and self.orbit.e is not None and self.orbit.is_ellipse():
            self._E = self._get_ascending_sign* np.arccos(((self.orbit.a - self._r) / (self.orbit.a * self.orbit.e)))
        else:
            return

        self.set_vars()

    def maneuver_vnc(self, delta_v):
        if self._dcm_rv is None:
            raise ValueError('Need DCM to Relate rth-vnc frames')

        print('Delta V (vnc): ' + str(delta_v))
        delta_v_rth = self._dcm_rv.dot(delta_v)
        return self.maneuver_rth(delta_v_rth)

    def maneuver_rth(self, delta_v):
        print('Delta V (rth): ' + str(delta_v))

        if self._dcm_er is None:
            raise ValueError('Need DCM to Relate eph-rth frames')

        delta_v_eph = self._dcm_er.transpose().dot(delta_v)

        return self.maneuver_eph(delta_v_eph)

    def maneuver_eph(self, delta_v):
        print('Delta V (eph): ' + str(delta_v))

        new_orbit = ConicOrbit(self.orbit.central_body)
        new_state = KeplarianState(new_orbit)
        new_state.r = self.r
        new_state.v = self.v - 0.5

        if self.dcm_ri is not None and self.dcm_er is not None:
            delta_v_xyz = self.dcm_ri.dot(self.dcm_er.dot(delta_v))
            new_state.v_xyz = self.v_xyz + delta_v_xyz
            new_state.r_xyz = self.r_xyz

            # TODO check sign here
            # new_state.set_fpa(self.fpa)
            # new_state.set_fpa(0)
            new_state.fpa = 0

            new_state.orbit.se_from_state(new_state)
            new_state.orbit.e_from_state(new_state)

            delta_omega = self._ta - new_state._ta
            new_state.orbit.arg_periapsis = self.orbit.arg_periapsis + delta_omega
            new_state.set_vars()
        return new_state

    def set_ascending(self):
        self._ascending = True

    def set_descending(self):
        self._ascending = False

    def _iteratively_find_E(self, tol=1e-12):
        return self._find_E_rec(self.M, self.M, tol=tol)

    def _find_E_rec(self, E, M, tol=1e-12):

        E1 = E - (E - self.orbit.e * math.sin(E) - M) / (1 - self.orbit.e * math.cos(E))
        dE = math.fabs(E - E1)

        if dE < tol:
            return E

        return self._find_E_rec(E1, M, tol=tol)

    def _is_ascending(self):
        if self._ascending is not None:
            return self._ascending

        angle_to_check = None

        if self._ta is not None:
            angle_to_check = self._ta

        elif self._E is not None:
            angle_to_check = self._E

        elif self._fpa is not None:
            angle_to_check = self._fpa

        return True if angle_to_check and 0 <= angle_to_check <= np.pi else False

    def _get_ascending_sign(self):
        return 1 if self._is_ascending() else -1

    def __repr__(self):

        x = ['%s Orbit State\n' % self.name]

        if self._r:
            x.append('r:     %0.4f km\n' % self._r)

        if self._r_rth is not None:
            x.append('r-rth:  %s km\n' % str(self._r_rth))

        if self._r_eph is not None:
            x.append('r-eph:  %s km\n' % str(self._r_eph))

        if self._r_xyz is not None:
            x.append('r-xyz:  %s km\n' % str(self._r_xyz))

        if self._v:
            x.append('v:     %0.4f km/s\n' % self._v)

        if self._v_rth is not None:
            x.append('v-rth:  %s km/s\n' % str(self._v_rth))

        if self._v_eph is not None:
            x.append('v-eph:  %s km/s\n' % str(self._v_eph))

        if self._v_xyz is not None:
            x.append('v-xyz:  %s km/s\n' % str(self._v_xyz))

        if self._ta is not None:
            x.append('ta:    %0.4f deg\n' % np.rad2deg(self._ta))
            x.append('ta:    %0.4f rad\n' % self._ta)

            if self._ta < 0:
                pos_ta = 2 * math.pi + self._ta
                x.append('ta:    %0.4f deg\n' % np.rad2deg(pos_ta))
                x.append('ta:    %0.4f rad\n' % pos_ta)

        if self._E is not None:
            x.append('E:     %0.4f deg\n' % np.rad2deg(self._E))
            x.append('E:     %0.4f rad\n' % self._E)

            if self._E < 0:
                pos_E = 2 * math.pi + self._E
                x.append('E:     %0.4f deg\n' % np.rad2deg(pos_E))
                x.append('E:     %0.4f rad\n' % pos_E)

        if self._M is not None:
            x.append('M:     %0.4f deg\n' % np.rad2deg(self._M))
            x.append('M:     %0.4f rad\n' % self._M)

        if self._fpa is not None:
            x.append('fpa:   %0.4f deg\n' % np.rad2deg(self._fpa))
            x.append('fpa:   %0.4f rad\n' % self._fpa)

        if self._theta is not None:
            x.append('theta: %0.4f deg\n' % np.rad2deg(self._theta))
            x.append('theta: %0.4f rad\n' % self._theta)

        if self._ttp is not None:
            x.append('ttp:   %0.4f s\n' % self._ttp)
            x.append('ttp:   %0.4f hr\n' % (self._ttp / 3600.0))

        if self._dcm_er is not None:
            x.append('C_er:  \n' + str(self._dcm_er) + '\n')

        if self._dcm_rv is not None:
            x.append('C_rv:  \n' + str(self._dcm_rv) + '\n')

        if self._dcm_ri is not None:
            x.append('C_ri:  \n' + str(self._dcm_ri) + '\n')

        return ''.join(x)


class ConicOrbit(object):

    def __init__(self, central_body, name=""):
        self.central_body = central_body
        self.name = name

        self._e = None
        self._a = None
        self._b = None
        self._rp = None
        self._ra = None
        self._p = None
        self._period = None
        self._se = None
        self._h = None
        self._h_xyz = None
        self._n = None
        self._asc_node = None
        self._arg_periapsis = None
        self._i = None
        self._vc = None

    def set_vars(self):
        self.e = None
        self.a = None
        self.b = None
        self.rp = None
        self.ra = None
        self.p = None
        self.period = None
        self.se = None
        self.h = None
        self.h_xyz = None
        self.n = None
        self.asc_node = None
        self.arg_periapsis = None
        self.i = None

    def set_circular(self):
        self.e = 0

        if self._a:
            self._p = self._a
            self._vc = np.sqrt(self.central_body.mu / self._a)

    @property
    def i(self):
        return self._i

    @i.setter
    def i(self, i=None):
        if self.i is not None:
            return

        if i is not None:
            self._i = i
        elif self.h_xyz is not None:
            self._i = np.arccos(self.h_xyz[2])
        else:
            return

        self.set_vars()

    @property
    def asc_node(self):
        return self._asc_node

    @asc_node.setter
    def asc_node(self, asc_node=None):
        if self._asc_node is not None:
            return

        if asc_node is not None:
            self._asc_node = asc_node
        elif self.h_xyz is not None and self.i is not None:
            sin_val = np.arcsin(self.h_xyz[0] / np.sin(self.i))
            cos_val = np.arccos(-self.h_xyz[1] / np.sin(self.i))
            self._asc_node = self._common_val(sin_val, cos_val)
        else:
            return

        self.set_vars()

    @property
    def arg_periapsis(self):
        return self._arg_periapsis

    @arg_periapsis.setter
    def arg_periapsis(self, arg_periapsis=None):
        if self._arg_periapsis is not None:
            return

        if arg_periapsis is not None:
            self._arg_periapsis = arg_periapsis
        else:
            return

        self.set_vars()

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, p=None):
        if self.p is not None:
            return

        if p:
            self._p = p

        elif self.a is not None and self.e is not None and self.is_ellipse():
            self._p = self.a * (1 - self.e ** 2)
        elif self.a is not None and self.e is not None and self.is_hyperbola():
            self._p = np.abs(self.a) * (self.e ** 2 - 1)
        elif self.h is not None:
            self._p = self.h ** 2 / self.central_body.mu
        else:
            return

        self.set_vars()

    @property
    def vc(self):
        return self._vc

    @vc.setter
    def vc(self, vc):
        self._vc = vc

    def get_rp(self):
        st = KeplarianState(self, "Periapsis")
        st.r = self.rp
        st.ta = 0.0
        st.set_ascending()
        return st

    def get_ra(self):
        st = KeplarianState(self, "Apoapsis")
        st.r = self.ra
        st.a = np.pi
        st.set_ascending()
        return st

    def is_circular(self):
        return self.e is not None and self._e == 0

    def is_ellipse(self):
        return (self.e is not None and self.e < 1.0) or (self._a is not None and self._a > 0)

    def is_hyperbola(self):
        return (self.e is not None and self.e > 1.0) or (self._a is not None and self._a < 0)

    def __repr__(self):
        x = ['%s Orbit Info\n' % self.name]

        if self._a is not None:
            x.append('a:      %0.4f km\n' % self._a)

        if self._b is not None:
            x.append('b:      %0.4f km\n' % self._b)

        if self.e is not None:
            x.append('e:      %0.4f\n' % self.e)

        if self._rp is not None:
            x.append('rp:     %0.4f km\n' % self._rp)

        if self._ra is not None:
            x.append('ra:     %0.4f km\n' % self._ra)

        if self._p is not None:
            x.append('p:      %0.4f km\n' % self._p)

        if self._period is not None:
            x.append('period: %0.4f s\n' \
                     'period: %0.4f hr\n' \
                     'period: %0.4f days\n' % (
                         self._period,
                         self._period / 3600,
                         self._period / 3600 / 24,
                     ))

        if self._h is not None:
            x.append('h:      %0.4f km^2/s\n' % self._h)

        if self._h_xyz is not None:
            x.append('h_xyz:  %s km^2/s\n' % str(self._h_xyz))

        if self._se is not None:
            x.append('se:     %0.4e km^2/s^2\n' % self._se)

        if self.vc is not None:
            x.append('vc:     %0.4f km/s\n' % self.vc)

        if self._i is not None:
            x.append('i:      %0.4f rad\n' % self._i)
            x.append('i:      %0.4f deg\n' % np.rad2deg(self._i))

        if self._asc_node is not None:
            x.append('Omega:  %0.4f rad\n' % self._asc_node)
            x.append('Omega:  %0.4f deg\n' % np.rad2deg(self._asc_node))

        if self._arg_periapsis is not None:
            x.append('omega:  %0.4f rad\n' % self._arg_periapsis)
            x.append('omega:  %0.4f deg\n' % np.rad2deg(self._arg_periapsis))

        return ''.join(x)


class LambertArc(object):
    def __init__(self, central_body, r1, r2, ta):
        self.r1 = r1
        self.r2 = r2

        self.ta = ta
        self.body = central_body

        self.c = conics_utils.loc(r1, r2, self.ta)
        self.s = 0.5 * (r1 + r2 + self.c)
        self.a_m = self.s / 2.
        self.alpha_m = 2 * np.arcsin(np.sqrt(self.s / (2 * self.a_m)))
        self.beta_m = 2 * np.arcsin(np.sqrt((self.s - self.c) / (2 * self.a_m)))
        self.TOF_m = np.sqrt(self.a_m ** 3 / self.body.mu) * \
                     (self.alpha_m - self.beta_m - np.sin(self.alpha_m) + np.sin(self.beta_m))
        self.p_m = (4. * self.a_m * (self.s - r1) * (self.s - r2)) / (self.c ** 2) * \
                   np.sin((self.alpha_m + self.beta_m) / 2.) ** 2

        self.transfer_orbit = ConicOrbit(self.body, 'Lambert Transfer')
        self.transfer_orbit.a = self.a_m
        self.transfer_orbit.p = self.p_m



########### Standard ###########

########### Local ###########
from orpytal.common import conics_utils
from orpytal.errors import ParameterUnavailableError

########### External ###########
import numpy as np


class Vector(object):
    def __init__(self, orbit, state, value, frame):
        self.orbit = orbit
        self.state = state

        self.value = value
        self.frame = frame
    @classmethod
    def from_vector(cls, orbit, state, vector):
        return cls(orbit, state, vector.value, vector.frame)

    def __str__(self):
        return '{} in {}'.format(self.value, self.frame.name)

    def __getitem__(self, i):
        return self.value.__getitem__(i)

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return self.value.__iter__()

    def __eq__(self, other):
        if isinstance(other, Vector):
            vec1 = self
            vec2 = other
            while vec1.frame != vec2.frame:
                try:
                    vec1 = vec1.to(vec2.frame)
                    break
                except ParameterUnavailableError:
                    pass

                try:
                    vec2 = vec2.to(vec1.frame)
                    break
                except ParameterUnavailableError:
                    raise ParameterUnavailableError(
                        'Tried to Compare [{}] to [{}] Values, but could not put them in the same frame'.format(self, other))

            return np.allclose(vec1.value, vec2.value)

        return NotImplemented

    def __neq__(self, other):
        if isinstance(other, Vector):
            return not self == other
        return NotImplemented

    def to(self, frame):
        if frame == InertialFrame:
            dcm = self.frame.inertial_dcm
        elif frame == RotatingFrame:
            dcm = self.frame.rotating_dcm
        elif frame == OrbitFixedFrame:
            dcm = self.frame.orbit_fixed_dcm
        else:
            raise ValueError('Frame ({}) is not recognized'.format(frame))

        if self.frame == frame:
            value = self.value
        else:
            value = dcm(self.orbit, self.state).dot(self.value) * self.value.units

        return Vector(self.orbit, self.state, value, frame)

    def inertial(self):
        return self.to(InertialFrame)

    def rotating(self):
        return self.to(RotatingFrame)

    def orbit_fixed(self):
        return self.to(OrbitFixedFrame)

    def norm(self):
        return np.linalg.norm(self.value) * self.value.units

    def unit(self):
        return (self.value / np.linalg.norm(self.value)).m

    def cross(self, vector2):
        assert self.frame == vector2.frame
        units = self.value.units * vector2.value.units
        return np.cross(self.value, vector2.value) * units

    def dot(self, vector2):
        if isinstance(vector2, Vector):
            assert self.frame == vector2.frame
            units = self.value.units * vector2.value.units
            val2 = vector2.value
        else:
            units = 1
            val2 = vector2

        return np.dot(self.value, val2) * units



class CoordinateFrame(object):
    name = None
    fn_name = None

    @classmethod
    def inertial_dcm(cls, orbit, state):
        raise NotImplementedError()

    @classmethod
    def rotating_dcm(cls, orbit, state):
        raise NotImplementedError()

    @classmethod
    def orbit_fixed_dcm(cls, orbit, state):
        raise NotImplementedError()

    @classmethod
    def vnc_dcm(cls, orbit, state):
        raise NotImplementedError()

    def __mul__(self, other):
        return (self, other)

    # def __rmul__(self, other):
    #     self.__mul__(other)

class RotatingFrame(CoordinateFrame):
    name = 'Rotating Frame'
    fn_name = 'rotating'

    @classmethod
    def inertial_dcm(cls, orbit, state):

        requirements = ['ascending_node', 'inclination', 'arg_latitude']
        if not conics_utils.state_orbit_satisfied(state, orbit, requirements):
            raise ParameterUnavailableError('Need ascending node, inclination and argument of latitude to convert to xyz')

        Omega = orbit.ascending_node
        theta = state.arg_latitude
        i = orbit.inclination
        dcm_ri = np.zeros((3, 3))
        dcm_ri[0, 0] = np.cos(Omega) * np.cos(theta) - np.sin(Omega) * np.cos(i) * np.sin(theta)
        dcm_ri[0, 1] = -np.cos(Omega) * np.sin(theta) - np.sin(Omega) * np.cos(i) * np.cos(theta)
        dcm_ri[0, 2] = np.sin(Omega) * np.sin(i)
        dcm_ri[1, 0] = np.sin(Omega) * np.cos(theta) + np.cos(Omega) * np.cos(i) * np.sin(theta)
        dcm_ri[1, 1] = -np.sin(Omega) * np.sin(theta) + np.cos(Omega) * np.cos(i) * np.cos(theta)
        dcm_ri[1, 2] = -np.cos(Omega) * np.sin(i)
        dcm_ri[2, 0] = np.sin(i) * np.sin(theta)
        dcm_ri[2, 1] = np.sin(i) * np.cos(theta)
        dcm_ri[2, 2] = np.cos(i)

        return dcm_ri

    @classmethod
    def rotating_dcm(cls, orbit, state):
        return np.eye(3)

    @classmethod
    def orbit_fixed_dcm(cls, orbit, state):
        return OrbitFixedFrame.rotating_dcm(orbit, state).transpose()

    @classmethod
    def vnc_dcm(cls, orbit, state):
        return np.array(
            [[np.sin(state.fpa), 0, np.cos(state.fpa)],
             [np.cos(state.fpa), 0, -np.sin(state.fpa)],
             [0, 1, 0]])


class VncFrame(CoordinateFrame):
    name = 'Vnc Frame'
    fn_name = 'vnc'

    @classmethod
    def inertial_dcm(cls, orbit, state):
        dcm_vr = cls.rotating_dcm(orbit, state)
        dcm_ri = RotatingFrame.inertial_dcm(orbit, state)
        return dcm_vr.dot(dcm_ri)

    @classmethod
    def rotating_dcm(cls, orbit, state):
        return RotatingFrame.vnc_dcm(orbit, state).transpose()

    @classmethod
    def orbit_fixed_dcm(cls, orbit, state):
        dcm_vr = cls.rotating_dcm(orbit, state)
        dcm_re = RotatingFrame.orbit_fixed_dcm(orbit, state)
        return dcm_vr.dot(dcm_re)

    @classmethod
    def vnc_dcm(cls, orbit, state):
        return np.eye(3)


class OrbitFixedFrame(CoordinateFrame):
    name = 'Orbit Fixed Frame'
    fn_name = 'orbit_fixed'

    @classmethod
    def inertial_dcm(cls, orbit, state):
        dcm_vr = cls.rotating_dcm(orbit, state)
        dcm_ri = RotatingFrame.inertial_dcm(orbit, state)
        return dcm_vr.dot(dcm_ri)

    @classmethod
    def rotating_dcm(cls, orbit, state):
        if state.ta is None:
            raise ParameterUnavailableError('Need true anomaly to convert to the rotating frame')
        return np.array(
            [[np.cos(state.ta), np.sin(state.ta), 0],
             [-np.sin(state.ta), np.cos(state.ta), 0],
             [0, 0, 1]]
        )

    @classmethod
    def orbit_fixed_dcm(cls, orbit, state):
        return np.eye(3)

    @classmethod
    def vnc_dcm(cls, orbit, state):
        dcm_vr = cls.rotating_dcm(orbit, state)
        dcm_rv = RotatingFrame.vnc_dcm(orbit, state)
        return dcm_vr.dot(dcm_rv)


class InertialFrame(CoordinateFrame):
    name = 'Inertial Frame'
    fn_name = 'inertial'

    @classmethod
    def inertial_dcm(cls, orbit, state):
        return np.eye(3)

    @classmethod
    def rotating_dcm(cls, orbit, state):
        return RotatingFrame.inertial_dcm(orbit, state).transpose()

    @classmethod
    def orbit_fixed_dcm(cls, orbit, state):
        return OrbitFixedFrame.inertial_dcm(orbit, state).transpose()

    @classmethod
    def vnc_dcm(cls, orbit, state):
        return VncFrame.inertial_dcm(orbit, state).transpose()



# Initialize Units
from pint import UnitRegistry, set_application_registry
from orpytal.common.conics_utils import orbit_setter, attribute_setter

import logging

print('Initializing Unit Registry')
units = UnitRegistry()
Q_ = units.Quantity
set_application_registry(units)

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


########### Standard ###########
import logging

########### Local ###########
from orpytal import KeplarianState, frames
from orpytal.common import units

########### External ###########
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np


def plot_orbit_fixed(orbit, **kwargs):
    return plot_orbit(orbit, frame='orbit_fixed', **kwargs)


def plot_orbit_inertial(orbit, **kwargs):
    return plot_orbit(orbit, frame='inertial', **kwargs)


def plot_orbit_inertial_3d(orbit, **kwargs):
    return plot_orbit(orbit, frame='inertial', planar=False, **kwargs)


def plot_orbit(orbit, frame=frames.OrbitFixedFrame.fn_name, planar=True, **kwargs):
    logging.info('Propagating and plotting {}'.format(orbit.name))

    start_st = KeplarianState(orbit)
    start_st.ta = 0 * units.rad
    traj = orbit.propagate_full_orbit()
    # traj = orbit.propagate_full_orbit(start_st)

    if hasattr(frame, 'fn_name'):
        frame = frame.fn_name

    x_list, y_list, z_list = getattr(traj, frame)()

    if 'label' not in kwargs:
        kwargs['label'] = orbit.name

    logging.info('Propagation Complete')

    if planar:
        plt.axis('equal')
        plt.plot(x_list, y_list, ls='dashed', **kwargs)
        plot_primary(orbit)
        plt.title(orbit.name)
        if frame == 'orbit_fixed':
            xaxis = 'e'
            yaxis = 'p'
        elif frame == 'inertial':
            xaxis = 'x'
            yaxis = 'y'
        else:
            return

        plt.xlabel("{} [{}]".format('$\hat{%s}$' % xaxis, 'km'))
        plt.ylabel("{} [{}]".format('$\hat{%s}$' % yaxis, 'km'))
        plt.legend()
    else:
        from orpytal import matlab_plotting as mplt
        mplt.init_cr3bp_3d_plot(1)
        mplt.title(orbit.name)
        mplt.default_3d_axes()
        mplt.plot_3d_traj(traj, label=orbit.name)
        mplt.show()


def plot_state_orbit_fixed(state):
    return plot_state(state, 'orbit_fixed')

def plot_state_inertial(state):
    return plot_state(state, 'inertial')

def plot_state(state, frame):
    if hasattr(frame, 'fn_name'):
        frame = frame.fn_name
    pos = getattr(state.position, frame)()
    return plt.plot(pos[0], pos[1],
                    marker='o',
                    c='r')


def plot_primary(orbit):
    fig = plt.gcf()
    ax = fig.gca()

    circle_primary = plt.Circle((0, 0), orbit.central_body.radius.m,
                                color='k',
                                fill=True,
                                lw=2.5)
    ax.add_artist(circle_primary)


def animate_orbit_fixed(orbit, step=0.01, **kwargs):
    ta_range = np.arange(0, 2 * np.pi, step)

    x_list = []
    y_list = []

    logging.info('Propagating... ')
    traj = []
    for ta in ta_range:
        st = KeplarianState(orbit)
        st._ta.value = ta * units.radians
        st._r.set(st, orbit)
        st._position.set(st, orbit)
        orbit_fixed_pos = st.position.orbit_fixed()
        x_list.append(orbit_fixed_pos[0])
        y_list.append(orbit_fixed_pos[1])
        traj.append(st)

    logging.info('Propagation Complete')
    traj_animator = TrajectoryAnimator(orbit, traj)
    traj_animator.animate()


class TrajectoryAnimator(object):
    def __init__(self,
                 orbit,
                 traj,
                 ta_fmt_str="ta = %0.4f [rad]"):

        self.traj = traj
        self.ta_fmt_str = ta_fmt_str

        plot_primary(orbit)
        plt.axis("equal")
        plt.title(orbit.name)
        plt.xlabel("{} [{}]".format('$\hat{e}$', 'km'))
        plt.ylabel("{} [{}]".format('$\hat{p}$', 'km'))
        ax = plt.gcf().gca()

        self.ta_text = ax.text(.05, 0.05, '',
                               fontsize=15,
                               horizontalalignment='left',
                               verticalalignment='bottom',
                               transform=plt.gca().transAxes)

        self.t_loc = ax.plot([], [], 'bo')[0]

        # For the axes....
        self.orbit_fixed_pos = [st.position.orbit_fixed() for st in self.traj]
        plt.plot([st[0] for st in self.orbit_fixed_pos],
                 [st[1] for st in self.orbit_fixed_pos],
                 c='0.85', ls='dashed')

    def _init(self):
        return self._animate(0)

    def _animate(self, i):
        st = self.traj[i]

        self.t_loc.set_data(self.orbit_fixed_pos[i][0], self.orbit_fixed_pos[i][1])

        if self.ta_fmt_str:
            self.ta_text.set_text(self.ta_fmt_str % self.traj[i].ta.m)

        return self.t_loc, self.ta_text

    def animate(self, save_as_gif_name=None, interval=20, fps=1):
        frames = np.arange(0, len(self.traj))
        ani = animation.FuncAnimation(plt.gcf(), self._animate, frames,
                                      interval=interval, blit=True)
        plt.rcParams['animation.ffmpeg_path'] = '/usr/bin/ffmpeg'

        if save_as_gif_name:
            print('Saving gif...')
            ani.save(save_as_gif_name, writer='imagemagick', fps=fps)
            print('Saving complete!')
        else:
            print('Animating...')
            plt.show()



########### Standard ###########

########### Local ###########
from orpytal.common import units, conics_utils
from orpytal import frames

########### External ###########
import numpy as np


class OrbitBase(object):
    symbol = ''

    def __init__(self, units):
        self.units = units

        self.requirements = None
        self._value = None
        self.evaluated = False

    def __str__(self):
        return '{}: {}'.format(self.symbol, self._value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value=None):
        if isinstance(value, units.Quantity):
            self._value = value.to(self.units)
        elif isinstance(value, frames.Vector):
            if isinstance(value.value, units.Quantity):
                value.value = value.value.to(self.units)
                self._value = value
            else:
                value.value = value.value * self.units
                self._value = value
        else:
            self._value = value * self.units

        # Resolve signs for units
        if self.units == units.rad:
            self._value = self._value % (2* np.pi)

        self.evaluated = True

    def __eq__(self, other):
        if isinstance(other, OrbitBase):
            return self._value == other._value
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, OrbitBase):
            return self._value != other._value
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, OrbitBase):
            return self._value > other._value
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, OrbitBase):
            return self._value >= other._value
        return NotImplemented

    def _lt__(self, other):
        if isinstance(other, OrbitBase):
            return self._value < other._value
        return NotImplemented

    def _e__(self, other):
        if isinstance(other, OrbitBase):
            return self._value <= other._value
        return NotImplemented

    def check_satisfied(self, obj, req):
        return conics_utils.check_satisfied(obj, req)

    def state_orbit_satisfied(self, state, orbit, requirements):
        return conics_utils.state_orbit_satisfied(state, orbit, requirements)


########### Standard ###########
import unittest

########### Local ###########
from orpytal import Orbit, KeplarianState, frames
from orpytal.common import units

########### External ###########
import numpy as np
from orpytal.planet_constants import BODIES_532, BODIES

earth = BODIES_532['EARTH']


class TestOrbit(unittest.TestCase):

    def test_ps3_problem_3(self):
        earth.radius = 6378.137 * units.km
        orbit = Orbit(earth, name='Test Orbit')
        state = KeplarianState(orbit, name='Test State')

        state.r = orbit.central_body.radius + 2000 * units.km
        vel_vector = np.array([-1.2, 6.7, 0.0]) * units.km / units.s
        state.velocity = vel_vector, frames.RotatingFrame

        print(orbit)
        print()
        print(state)

        from orpytal import plotting
        from matplotlib import pyplot as plt
        plotting.plot_orbit(orbit, frames.OrbitFixedFrame)
        plotting.plot_state(state, frames.OrbitFixedFrame)
        plt.show(block=False)

        # Assert Orbit Quantities
        assert (np.isclose(orbit.p, 7891.62633))
        assert (np.isclose(orbit.e, 0.17829470376))
        assert (np.isclose(orbit.h, 56085.7))
        assert (np.isclose(orbit.a, 8150.7299))
        assert (np.isclose(orbit.b, 7648.4885248))
        assert (np.isclose(orbit.period, 7323.281580402567))
        assert (np.isclose(orbit.n, 0.0008579740159))
        assert (np.isclose(orbit.se, -24.451824967))
        assert (np.isclose(orbit.rp, 9603.961874))
        assert (np.isclose(orbit.ra, 6697.497928))

        # Assert State Quantities
        assert (np.isclose(state.ta, -1.8977792920))
        assert (np.isclose(state.v, 6.80661443))
        assert (np.isclose(state.fpa, -0.17722538))
        assert (np.isclose(state.t_since_rp, -1802.7593))
        assert (np.isclose(state.M, -1.546720643))
        assert (np.isclose(state.E, -1.7229553535))

        pass

    def test_ps4_problem_4(self):
        earth.radius = 6378.137 * units.km
        orbit = Orbit(earth, name='Test Orbit')
        state = KeplarianState(orbit, name='Test State')
        orbit.a = 10 * earth.radius
        orbit.e = 0.4
        orbit.i = 30 * units.deg
        orbit.ascending_node = 45 * units.deg
        orbit.arg_periapsis = -90 * units.deg
        state.ta = 135 * units.deg

        assert (np.isclose(state.r, 74707))
        assert (np.isclose(state.v, 2.10276))
        assert (np.isclose(orbit.h, 146135))
        assert (np.isclose(state.fpa.to('deg'), 21.52399*units.deg))
        assert (np.isclose(state.E.to('deg'), 115.355*units.deg))


        pass


if __name__ == '__main__':
    unittest.main()



########### Standard ###########
import unittest

########### Local ###########
from orpytal import Orbit, KeplarianState,frames
from orpytal.common import units
from orpytal.planet_constants import BODIES_532

########### External ###########
import numpy as np

earth = BODIES_532['EARTH']
earth.radius = 6378.137 * units.km

class TestOrbit(unittest.TestCase):

    def test_JS_3Dex1(self):
        orbit = Orbit(earth, name='Test Orbit')
        state = KeplarianState(orbit, name='Test State')

        orbit.a = 8 * earth.radius
        orbit.e = 0.7
        orbit.i = 30 * units.deg
        orbit.ascending_node = 60 * units.deg
        orbit.arg_periapsis = 90 * units.deg
        state.ta = 90 * units.deg

        state.position.inertial()
        pass

    def test_Iex(self):
        orbit = Orbit(earth, name='Test Orbit')
        state = KeplarianState(orbit, name='Test State')

        state.position = frames.Vector(orbit, state, np.array([1.6772, -1.6722, 2.3719]) * earth.radius, frames.InertialFrame)
        state.velocity = frames.Vector(orbit, state, np.array([3.1574, 2.4987, 0.4658]), frames.InertialFrame)


        pass


if __name__ == '__main__':
    unittest.main()



########### Standard ###########
import logging

########### Local ###########
import orpytal.orbit
import orpytal.planet_constants
from orpytal.common import units

########### External ###########
import matlab.engine

logging.info("Starting Matlab Engine....")
eng = matlab.engine.start_matlab()
logging.info("Matlab Engine Started!")


def eval(cmd, nargout=0):
    eng.eval(cmd, nargout=nargout)


def figure(fig_number=None):
    if fig_number:
        eng.figure(float(fig_number))
    else:
        eng.figure()


def init_cr3bp_3d_plot(fig_number=None):
    figure(fig_number)

    eval("hold on;")
    eval("axis equal;")
    default_3d_axes()


def plot_sphere(x, y, z, radius):
    if isinstance(radius, units.Quantity):
        radius = radius.to('km').m
    eng.workspace['radius'] = radius
    eng.workspace['x'] = float(x)
    eng.workspace['y'] = float(y)
    eng.workspace['z'] = float(z)
    eval("[sx sy sz] = sphere;")
    eval("surf(sx*radius + x, sy*radius + y, sz*radius + z, 'FaceColor', [0 0 0]);")


def plot_primary(primary):
    if isinstance(primary, orbit.Orbit):
        radius = primary.central_body.radius
    elif isinstance(primary, planet_constants.CentralBody):
        radius = primary.radius
    else:
        radius = primary
    return plot_sphere(0, 0, 0, radius)


def plot_p1(system):
    return plot_sphere(-system.mu, 0, 0, system.primary_body.radius / system.lstar)


def legend(*args):
    if not args:
        eng.legend()
    else:
        eng.legend(*args)


def _prepare_plot(x_vals, y_vals, z_vals, **kwargs):
    x_vals = _to_list(x_vals)
    y_vals = _to_list(y_vals)
    z_vals = _to_list(z_vals)

    x_vals = matlab.double(x_vals)
    y_vals = matlab.double(y_vals)
    z_vals = matlab.double(z_vals)
    args = format_args(**kwargs)
    return x_vals, y_vals, z_vals, args


def _to_list(val):
    if not hasattr(val, '__len__'):
        val = [val]
    return val


def scatter3(x_vals, y_vals, z_vals, **kwargs):
    x_vals, y_vals, z_vals, args = _prepare_plot(x_vals, y_vals, z_vals, **kwargs)

    args = replace_arg_key(args, 'Color', 'MarkerEdgeColor')

    return eng.scatter3(x_vals, y_vals, z_vals, *args, nargout=1)


def replace_arg_key(args, key, new_key):
    args_list = list(args)
    for i, arg in enumerate(args_list):
        if arg == key:
            args_list[i] = new_key
    return tuple(args_list)


def format_args(**kwargs):
    args = ()
    if 'c' in kwargs:
        color = kwargs['c']
        if isinstance(color, str):
            pass  # no op
        elif hasattr(color, '__len__'):
            color = matlab.double(list(color))[0]
        else:
            raise TypeError('need list or string for color')

        args = args + ("Color", color)

    if 'ls' in kwargs:
        ls = kwargs['ls']
        if ls == 'dashed':
            ls = '--'
        elif ls == 'dotted':
            ls = ':'
        args = args + ('LineStyle', ls)

    if 'label' in kwargs:
        args = args + ('DisplayName', kwargs['label'])

    if 'marker' in kwargs:
        args = args + ("Marker", kwargs['marker'])

    return args


def plot3(x_vals, y_vals, z_vals, **kwargs):
    x_vals, y_vals, z_vals, args = _prepare_plot(x_vals, y_vals, z_vals, **kwargs)
    return eng.plot3(x_vals, y_vals, z_vals, *args, nargout=1)


def scatter3_states(states, **kwargs):
    return scatter3(
        [st.x for st in states],
        [st.y for st in states],
        [st.z for st in states],
        **kwargs
    )


def plot_3d_traj_list(traj_list, **kwargs):
    logging.debug("Plotting list of 3D Trajectories....")
    for i, traj in enumerate(traj_list):
        if (i + 1) % 10 == 0:
            logging.info("Traj {} / {}".format(i + 1, len(traj_list)))
        plot_3d_traj(traj, **kwargs)
    logging.debug("Plotting list of 3D Trajectories done!")


def plot_3d_traj(traj, **kwargs):
    if 'color' in traj.metadata and 'c' not in kwargs:
        kwargs['c'] = traj.metadata['color']

    xs, ys, zs = traj.inertial()
    return plot3(xs, ys, zs, **kwargs)


def title(title, latex=True):
    args = ()
    if latex:
        args = args + ("Interpreter", 'Latex')
    return eng.title(title, *args)


def xlabel(xlabel, latex=True):
    args = ()
    if latex:
        args = args + ("Interpreter", 'Latex')

    return eng.xlabel(xlabel, *args)


def ylabel(ylabel, latex=True):
    args = ()
    if latex:
        args = args + ("Interpreter", 'Latex')

    return eng.ylabel(ylabel, *args)


def zlabel(zlabel, latex=True):
    args = ()
    if latex:
        args = args + ("Interpreter", 'Latex')

    return eng.zlabel(zlabel, *args)


def default_3d_axes():
    unit = 'km'
    fmt_str = '$\hat{%s}$ [%s]'

    xlabel(fmt_str % ('x', unit), latex=True)
    ylabel(fmt_str % ('y', unit), latex=True)
    zlabel(fmt_str % ('z', unit), latex=True)


def grid():
    eng.eval("grid on;", nargout=0)


def show():
    pass



##########################################################
# Reference: https://wiki.kerbalspaceprogram.com/wiki/Category:Celestials
##########################################################


########### Standard ###########

########### Local ###########
from common import units
from planet_constants import CentralBody

########### External ###########

##########################################################
# Values
##########################################################

G = 6.67408e-11 / (1000 ** 3) * units.km ** 3 / (units.kg * units.s ** 2)

##########################################################
# Planets
##########################################################

mu_units = (units.km ** 3 / units.s ** 2)

BODIES = {}

################## Sun ##################

BODIES['KERBOL'] = CentralBody(
    name='Kerbol',
    radius=26160000 * units.m,
    mu=1.1723328e18 * units('m^3/s^2')
)

################## Moho ##################

BODIES['MOHO'] = CentralBody(
    name='Moho',
    radius=250000 * units.m,
    mu=1.6860938e11 * units('m^3/s^2'),
    a=5263138304 * units.m,
    period=5657995 * units.s,
    e=0.2 * units.dimensionless,
    i=7 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=15 * units.deg,
    ascending_node=70 * units.deg

)
################## Eve ##################

BODIES['EVE'] = CentralBody(
    name='Eve',
    radius=700000 * units.m,
    mu=8.1717302e12 * units('m^3/s^2'),
    a=9832684544 * units.m,
    period=5657995 * units.s,
    e=0.01 * units.dimensionless,
    i=2.1 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=15 * units.deg
)

BODIES['GILLY'] = CentralBody(
    name='Gilly',
    radius=31500000 * units.m,
    mu=8289449.8 * units('m^3/s^2'),
    a=31500000 * units.m,
    period=388587 * units.s,
    e=0.55 * units.dimensionless,
    i=12 * units.degrees,
    parent=BODIES['EVE'],
    arg_periapsis=10 * units.deg,
    ascending_node=80 * units.deg
)

################## Kerbin ##################

BODIES['KERBIN'] = CentralBody(
    name='Kerbin',
    radius=600000 * units.m,
    mu=3.5316000e12 * units('m^3/s^2'),
    a=13599840256 * units.m,
    period=9203545 * units.s,
    e=0 * units.dimensionless,
    i=0 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)

BODIES['MUN'] = CentralBody(
    name='Mun',
    radius=200000 * units.m,
    mu=6.5138398e10 * units('m^3/s^2'),
    a=12000000 * units.m,
    period=138984 * units.s,
    e=0 * units.dimensionless,
    i=0 * units.degrees,
    parent=BODIES['KERBIN'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)

BODIES['MINIMUS'] = CentralBody(
    name='Minimus',
    radius=60000 * units.m,
    mu=1.7658000e9 * units('m^3/s^2'),
    a=47000000 * units.m,
    period=1077311 * units.s,
    e=0 * units.dimensionless,
    i=6 * units.degrees,
    parent=BODIES['KERBIN'],
    arg_periapsis=38 * units.deg,
    ascending_node=78 * units.deg
)

################## Duna ##################

BODIES['DUNA'] = CentralBody(
    name='Duna',
    radius=320000 * units.m,
    mu=3.0136321e11 * units('m^3/s^2'),
    a=20726155264 * units.m,
    period=17315400 * units.s,
    e=0.051 * units.dimensionless,
    i=0.06 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=135.5 * units.deg
)

BODIES['IKE'] = CentralBody(
    name='Ike',
    radius=130000 * units.m,
    mu=1.8568369e10 * units('m^3/s^2'),
    a=3200000 * units.m,
    period=65518 * units.s,
    e=0.03 * units.dimensionless,
    i=0.2 * units.degrees,
    parent=BODIES['DUNA'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)

################## Dres ##################

BODIES['DRES'] = CentralBody(
    name='Dres',
    radius=138000 * units.m,
    mu=2.1484489e10 * units('m^3/s^2'),
    a=40839348203 * units.m,
    period=47893063 * units.s,
    e=0.145 * units.dimensionless,
    i=5 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=90 * units.deg,
    ascending_node=280 * units.deg
)

################## Jool ##################

BODIES['JOOL'] = CentralBody(
    name='Jool',
    radius=6000000 * units.m,
    mu=2.8252800e14 * units('m^3/s^2'),
    a=68773560320 * units.m,
    period=104661432 * units.s,
    e=0.05 * units.dimensionless,
    i=1.304 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=52 * units.deg
)

BODIES['LAYTHE'] = CentralBody(
    name='Laythe',
    radius=500000 * units.m,
    mu=1.9620000e12 * units('m^3/s^2'),
    a=27184000 * units.m,
    period=52981 * units.s,
    e=0 * units.dimensionless,
    i=0 * units.deg,
    parent=BODIES['JOOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)
BODIES['VALL'] = CentralBody(
    name='Vall',
    radius=300000 * units.m,
    mu=2.0748150e11 * units('m^3/s^2'),
    a=43152000 * units.m,
    period=105962 * units.s,
    e=0 * units.dimensionless,
    i=0 * units.deg,
    parent=BODIES['JOOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)
BODIES['TYLO'] = CentralBody(
    name='Jool',
    radius=600000 * units.m,
    mu=2.8252800e11 * units('m^3/s^2'),
    a=68500000 * units.m,
    period=211926 * units.s,
    e=0 * units.dimensionless,
    i=0.025 * units.deg,
    parent=BODIES['JOOL'],
    arg_periapsis=0 * units.deg,
    ascending_node=0 * units.deg
)
BODIES['BOP'] = CentralBody(
    name='Bop',
    radius=65000 * units.m,
    mu=2.4868349e9 * units('m^3/s^2'),
    a=128500000 * units.m,
    period=544507 * units.s,
    e=0.235 * units.dimensionless,
    i=15 * units.deg,
    parent=BODIES['JOOL'],
    arg_periapsis=25 * units.deg,
    ascending_node=10 * units.deg
)
BODIES['POL'] = CentralBody(
    name='Pol',
    radius=179890000 * units.m,
    mu=7.2170208e8 * units('m^3/s^2'),
    a=179890000 * units.m,
    period=901903 * units.s,
    e=0.171 * units.dimensionless,
    i=4.25 * units.deg,
    parent=BODIES['JOOL'],
    arg_periapsis=15 * units.deg,
    ascending_node=2 * units.deg
)

################## Eeloo ##################

BODIES['EELOO'] = CentralBody(
    name='Eeloo',
    radius=210000 * units.m,
    mu=7.4410815e10 * units('m^3/s^2'),
    a=90118820000 * units.m,
    period=156992048 * units.s,
    e=0.26 * units.dimensionless,
    i=6.15 * units.deg,
    parent=BODIES['KERBOL'],
    arg_periapsis=260 * units.deg,
    ascending_node=50 * units.deg
)

################## Easier Access ##################


kerbol = BODIES['KERBOL']

moho = BODIES['MOHO']

eve = BODIES['EVE']
gilly = BODIES['GILLY']

kerbin = BODIES['KERBIN']
mun = BODIES['MUN']
minimus = BODIES['MINIMUS']

duna = BODIES['DUNA']
ike = BODIES['IKE']

dres = BODIES['DRES']

jool = BODIES['JOOL']
laythe = BODIES['LAYTHE']
vall = BODIES['VALL']
tylo = BODIES['TYLO']
bop = BODIES['BOP']
pol = BODIES['POL']

eeloo = BODIES['EELOO']



import logging
import numpy as np

from orpytal import frames
from orpytal.errors import ParameterUnavailableError

def loc(v1, v2, theta):
    return np.sqrt(v1 ** 2 + v2 ** 2 - 2 * v1 * v2 * np.cos(theta))


def loc_angle(v1, v2, v3):
    return np.arccos((v3 ** 2 - v1 ** 2 - v2 ** 2) / -(2 * v1 * v2))

def common_val(sin_val, cos_val, tol=1e-3):
    one = sin_val
    if one > 0:
        two = np.pi - sin_val
    else:
        two = -np.pi - sin_val

    three = cos_val
    four = -cos_val

    if np.abs(one - three) < tol or np.abs(one - four) < tol:
        val = one
    elif np.abs(two - three) < tol or np.abs(two - four) < tol:
        val = two
    else:
        raise ValueError('No common value found for %f and %f\n' % (sin_val, cos_val))

    return val


def orbit_setter(setter_function):
    def wrapper(*args):
        orbit_value = args[0]

        # If this orbital parameter already has a value, return immediately
        if orbit_value.evaluated:
            return False

        value_before = orbit_value.value
        try:
            setter_function(*args)
        except ParameterUnavailableError as e:
            logging.debug('Assertion Error: {}'.format(e))
            pass

        value_after = orbit_value.value
        if value_before != value_after:
            logging.debug('Set {} to {}'.format(orbit_value.symbol, orbit_value.value))


        # Return true if the value of the parameter has changed as as result of the function call
        return value_before != value_after

    return wrapper

def attribute_setter(setter_function):
    def wrapper(*args):
        orbit_or_state = args[0]
        val = args[1]
        var_name = setter_function.__name__

        if isinstance(val, tuple) and \
                orbit_or_state.__class__.__name__ == 'KeplarianState' and \
                hasattr(val[0], '__len__') and \
                (isinstance(val[1], frames.CoordinateFrame) or val[1].__bases__[0] == frames.CoordinateFrame):
            val = frames.Vector(orbit_or_state.orbit, orbit_or_state, val[0], val[1])

        setter_function(orbit_or_state, val)
        logging.debug('Set {} to {}'.format(var_name, val))

        orbit_or_state.set_vars()

    return wrapper


def check_satisfied(obj, req):
    return hasattr(obj, '_' + req) and getattr(obj, '_' + req).evaluated

def state_orbit_satisfied(state, orbit, requirements):
    if isinstance(requirements, str):
        return check_satisfied(state, requirements) or check_satisfied(orbit, requirements)
    else:
        return all([check_satisfied(state, req) or check_satisfied(orbit, req) for req in requirements])


from orpytal.ksp import ksp_planet_constants
from orpytal import plotting
from matplotlib import pyplot as plt
from orpytal import matlab_plotting as mplt

kerbin = ksp_planet_constants.kerbin
mun = ksp_planet_constants.mun

plt.figure(1)
plotting.plot_orbit_inertial(ksp_planet_constants.moho.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.eve.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.kerbin.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.duna.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.dres.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.jool.orbit)
plotting.plot_orbit_inertial(ksp_planet_constants.eeloo.orbit)
plt.title('Kerbal Planets')
plt.show(block=False)

plotting.plot_orbit_inertial_3d(ksp_planet_constants.mun.orbit)
plotting.plot_orbit_inertial_3d(ksp_planet_constants.minimus.orbit)
mplt.title('Kerbin Moons')
mplt.plot_primary(ksp_planet_constants.kerbin)
mplt.legend()

# plt.figure(2)
# plotting.plot_orbit_inertial(ksp_planet_constants.laythe.orbit)
# plotting.plot_orbit_inertial(ksp_planet_constants.vall.orbit)
# plotting.plot_orbit_inertial(ksp_planet_constants.tylo.orbit)
# plotting.plot_orbit_inertial(ksp_planet_constants.bop.orbit)
# plotting.plot_orbit_inertial(ksp_planet_constants.pol.orbit)
# plt.title('Moons of Jool')
# plt.show(block=False)



########### Standard ###########

########### Local ###########
from orpytal import Orbit, bodies, plotting
from orpytal.common import units

########### External ###########
import numpy as np

np.set_printoptions(precision=4)

orbit = Orbit(bodies.earth, name='Test Orbit', a=37800*units.km, p=16000*units.km)
plotting.plot_orbit(orbit)

print(orbit)


from orpytal import frames

class Trajectory(object):
    name = None

    def __init__(self, states, name=None):
        self.states = states
        self.t_range = [self.states[0].t_since_rp, self.states[-1].t_since_rp]
        self.times = [st.t_since_rp for st in self.states]
        self.metadata = {}
        self.name = name
        self.frame = states[0].position.frame

    def x_vals(self, frame='orbit_fixed'):
        x_vals = []
        for st in self.states:
            frame_fn = getattr(st.position, frame)
            x_vals.append(frame_fn()[0])

        return x_vals

    def vals(self, frame):
        x_vals = []
        y_vals = []
        z_vals = []

        for st in self.states:
            frame_fn = getattr(st.position, frame)
            pos = frame_fn().value
            x_vals.append(pos[0].m)
            y_vals.append(pos[1].m)
            z_vals.append(pos[2].m)

        return x_vals, y_vals, z_vals

    def inertial(self):
        self.frame = frames.InertialFrame
        return self.in_frame('inertial')

    def orbit_fixed(self):
        self.frame = frames.OrbitFixedFrame
        return self.in_frame('orbit_fixed')

    def in_frame(self, frame_name):
        self.x_vals, self.y_vals, self.z_vals = self.vals(frame_name)
        return self.x_vals, self.y_vals, self.z_vals

    def start(self):
        return self.states[0]

    def end(self):
        return self.states[-1]



name = "orpytal"

from orpytal.orbit import Orbit
from orpytal.state import KeplarianState
from orpytal.common import units, Q_
import orpytal.planet_constants as bodies

def get_body(body_name):
    return bodies[str(body_name).lower()]


########### Standard ###########
import logging

########### Local ###########
from orpytal.base import OrbitBase
from orpytal.common import units, orbit_setter, attribute_setter
from orpytal.errors import ParameterUnavailableError
from orpytal import frames
from orpytal.state import KeplarianState
from orpytal.trajectory import Trajectory

########### External ###########
import numpy as np


class Orbit(object):
    def __init__(self, central_body, name='', **kwargs):
        self.central_body = central_body
        self.name = name

        self._p = SemiLatusRectum()
        self._e = Eccentricity()
        self._e_vec = EccentricityVector()
        self._h = AngularMomentumMagnitude()
        self._angular_momentum = AngularMomentumVector()

        # Orientation angles
        self._arg_periapsis = ArgumentOfPeriapsis()
        self._ascending_node = LongitudeOfAscendingNode()
        self._inclination = Inclination()
        self._ascending_node_vec = AscendingNodeVector()
        self._i = self._inclination

        self._a = SemimajorAxis()
        self._b = SemiminorAxis()
        self._period = OrbitalPeriod()
        self._n = MeanMotion()
        self._se = SpecificEnergy()
        self._rp = Periapsis()
        self._ra = Apoapsis()

        self.vars = [
            self._p,
            self._e,
            self._e_vec,
            self._h,
            self._angular_momentum,
            self._a,
            self._b,
            self._period,
            self._n,
            self._se,
            self._rp,
            self._ra,
            self._arg_periapsis,
            self._ascending_node,
            self._ascending_node_vec,
            self._inclination
        ]

        for k, v in kwargs.items():
            try:
                attribute = getattr(self, '_' + k)
                attribute.value = v
            except AttributeError as e:
                logging.error('Orbit has no parameter named {}.'.format(k))
                raise e

        self.set_vars()

    def __str__(self):
        x = ['%s Orbit Info' % self.name]
        if self.e is not None:
            x.append('Circular: {}'.format(self.circular()))
        if self.i is not None:
            x.append('Equitorial: {}'.format(self.equitorial()))
        for var in self.vars:
            if var.evaluated:
                x.append(str(var))

        return '\n'.join(x)

    def propagate_full_orbit(self, n=150):
        # t_range = list(np.arange(0, self.period.m, step))
        t_range = np.linspace(start=0, stop=self.period.m, num=n)


        st_list = []
        for t in t_range:
            st = KeplarianState(self)
            logging.warning('-------------------------')
            st.t_since_rp = t
            st._t_since_rp.value = t
            st._M.set(st, self)
            st._E.set(st, self)
            st._r.set(st, self)
            st._ta.set(st, self)
            st._arg_latitude.set(st, self)
            st._position.set(st, self)
            st._velocity.set(st, self)

            st_list.append(st)

        return Trajectory(st_list)

    def propagate_full_orbit_ta(self, state, step=0.1):
        ta_range = list(np.arange(state.ta, state.ta + 2 * np.pi, step))
        ta_range.append(state.ta.m + 2 * np.pi)

        st_list = []

        for ta in ta_range:
            st = KeplarianState(self)
            st._ta.value = ta
            st._r.set(st, self)
            st._position.set(st, self)
            st._arg_latitude.set(st, self)
            st._t_since_rp.set(st, self)
            st_list.append(st)

        return Trajectory(st_list)

    def set_vars(self):
        for var in self.vars:
            new_value_set = var.set(self)
            if new_value_set:
                logging.debug('Set {} to {}'.format(var.symbol, var.value))
                self.set_vars()
                break

    def from_state(self, state):
        for var in self.vars:
            new_value_set = var.set_from_state(state, self)
            if new_value_set:
                self.set_vars()
                self.from_state(state)
                return True

    def equitorial(self, tol=1e-8):
        if self.inclination is not None:
            return np.abs(self.inclination) < tol
        else:
            raise ParameterUnavailableError('Need inclination to see if equitorial')

    def circular(self, tol=1e-8):
        if self.e is not None:
            return self.e < tol
        else:
            raise ParameterUnavailableError('Need eccentricity to see if circular')

    def compare(self, orbit):
        # Compare which vars are evaluated
        evaluated_vars = sorted([v.symbol for v in self.vars if v.evaluated])
        other_evaluated_vars = sorted([v.symbol for v in orbit.vars if v.evaluated])
        for i, var in enumerate(evaluated_vars):
            assert var == other_evaluated_vars[i]

        # Compare values of evaluated variables
        for var in self.vars:
            if var.evaluated and hasattr(orbit, var.symbol):
                if isinstance(var.value, units.Quantity):
                    same = np.isclose(var.value, getattr(orbit, var.symbol))
                elif isinstance(var.value, frames.Vector):

                    vec1 = var.value
                    vec2 = getattr(orbit, var.symbol)
                    same = vec1 == vec2
                else:
                    same = False

            if same:
                logging.debug('Checked {} [✓]'.format(var.symbol))
            else:
                logging.warning('Error Found for {} [x]'.format(var.symbol))
                logging.warning('My value: {}'.format(var.value))
                logging.warning('Their value: {}'.format(getattr(orbit, var.symbol)))


    @property
    def a(self):
        return self._a.value

    @a.setter
    @attribute_setter
    def a(self, a):
        self._a.value = a

    @property
    def b(self):
        return self._b.value

    @b.setter
    @attribute_setter
    def b(self, b):
        self._b.value = b

    @property
    def period(self):
        return self._period.value

    @period.setter
    @attribute_setter
    def period(self, period):
        self._period.value = period

    @property
    def n(self):
        return self._n.value

    @n.setter
    @attribute_setter
    def n(self, n):
        self._n.value = n

    @property
    def se(self):
        return self._se.value

    @se.setter
    @attribute_setter
    def se(self, se):
        self._se.value = se

    @property
    def rp(self):
        return self._rp.value

    @rp.setter
    @attribute_setter
    def rp(self, rp):
        self._rp.value = rp

    @property
    def ra(self):
        return self._ra.value

    @ra.setter
    @attribute_setter
    def ra(self, ra):
        self._ra.value = ra

    @property
    def e(self):
        return self._e.value

    @e.setter
    @attribute_setter
    def e(self, e):
        self._e.value = e

    @property
    def e_vec(self):
        return self._e_vec.value

    @e_vec.setter
    @attribute_setter
    def e_vec(self, e_vec):
        self._e_vec.value = e_vec

    @property
    def p(self):
        return self._p.value

    @p.setter
    @attribute_setter
    def p(self, p):
        self._p.value = p

    @property
    def h(self):
        return self._h.value

    @h.setter
    @attribute_setter
    def h(self, h):
        self._h.value = h

    @property
    def angular_momentum(self):
        return self._angular_momentum.value

    @angular_momentum.setter
    @attribute_setter
    def angular_momentum(self, angular_momentum):
        self._angular_momentum.value = angular_momentum

    @property
    def arg_periapsis(self):
        return self._arg_periapsis.value

    @arg_periapsis.setter
    @attribute_setter
    def arg_periapsis(self, arg_periapsis):
        self._arg_periapsis.value = arg_periapsis

    @property
    def ascending_node(self):
        return self._ascending_node.value

    @ascending_node.setter
    @attribute_setter
    def ascending_node(self, ascending_node):
        self._ascending_node.value = ascending_node

    @property
    def ascending_node_vec(self):
        return self._ascending_node_vec.value

    @ascending_node_vec.setter
    @attribute_setter
    def ascending_node_vec(self, ascending_node_vec):
        self._ascending_node_vec.value = ascending_node_vec

    @property
    def i(self):
        return self.inclination

    @i.setter
    def i(self, i):
        self.inclination = i

    @property
    def inclination(self):
        return self._inclination.value

    @inclination.setter
    @attribute_setter
    def inclination(self, inclination):
        self._inclination.value = inclination


class OrbitValue(OrbitBase):
    def set(self, orbit):
        raise NotImplementedError()

    def set_from_state(self, state, orbit):
        return False

    def satisfied(self, orbit, requirements):
        if isinstance(requirements, str):
            return self.check_satisfied(orbit, requirements)
        else:
            return all([self.check_satisfied(orbit, req) for req in requirements])


class SemiLatusRectum(OrbitValue):
    symbol = 'p'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('a', 'e'),
            ('h')
        ]
        self.orbit_state_requirements = [
            ('r', 'e', 'ta')
        ]

    @orbit_setter
    def set(self, orbit):

        # a(1-e^2)
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = orbit.a * (1 - orbit.e ** 2)

        # h^2/mu
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = orbit.h ** 2 / orbit.central_body.mu

    @orbit_setter
    def set_from_state(self, state, orbit):

        # r(1+ecos(ta))
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            self.value = state.r * (1 + orbit.e * np.cos(state.ta))


class Eccentricity(OrbitValue):
    symbol = 'e'

    def __init__(self):
        super().__init__(units.dimensionless)
        self.orbit_requirements = [
            ('se', 'h'),
            ('a', 'rp'),
            ('e_vec')
        ]
        self.orbit_state_requirements = [
            ('r', 'v', 'fpa')
        ]

    @orbit_setter
    def set(self, orbit):

        # sqrt(1 + (2 se h^2) / (mu ^ 2) )
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = np.sqrt(1 + (2 * orbit.se * orbit.h ** 2) / (orbit.central_body.mu ** 2))

        # 1 - rp/a
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = 1.0 - orbit.rp / orbit.a

        # |e|
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = np.linalg.norm(orbit.e_vec)

    @orbit_setter
    def set_from_state(self, state, orbit):

        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            self.value = np.sqrt(((state.r * state.v ** 2) / orbit.central_body.mu - 1) ** 2 * np.cos(state.fpa) ** 2 + \
                                 np.sin(state.fpa) ** 2)


class EccentricityVector(OrbitValue):
    symbol = 'e_vec'

    def __init__(self):
        super().__init__(units.dimensionless)
        self.orbit_requirements = [
        ]
        self.orbit_state_requirements = [
            ('position', 'velocity')
        ]

    @orbit_setter
    def set(self, orbit):
        return

    @orbit_setter
    def set_from_state(self, state, orbit):
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            v = state.velocity.inertial()  # inertial velocity (xyz)
            r = state.position.inertial()  # inertial position (xyz)
            mu = orbit.central_body.mu  # gravitational parameter
            val = ((v.dot(v) - mu / r.norm()) * r.value - r.dot(v) * v.value) / mu
            self.value = frames.Vector(orbit, state, val, frames.InertialFrame)


class AngularMomentumMagnitude(OrbitValue):
    symbol = 'h'

    def __init__(self):
        super().__init__(units.km ** 2 / units.s)
        self.orbit_requirements = [
            ('p'),
            ('angular_momentum')
        ]
        self.orbit_state_requirements = [
            ('position', 'velocity')
        ]

    @orbit_setter
    def set(self, orbit):

        # sqrt(p mu)
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = np.sqrt(orbit.p * orbit.central_body.mu)
        # | h |
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = np.linalg.norm(orbit.angular_momentum)

    @orbit_setter
    def set_from_state(self, state, orbit):

        # | r cross v |
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            self.value = np.linalg.norm(np.cross(state.position.value, state.velocity.value))


class AngularMomentumVector(OrbitValue):
    symbol = 'angular_momentum'

    def __init__(self):
        super().__init__(units.km ** 2 / units.s)
        self.orbit_requirements = [
        ]
        self.orbit_state_requirements = [
            ('ascending_node', 'inclination', 'h'),
            ('position', 'velocity')
        ]

    @orbit_setter
    def set(self, orbit):
        return False

    @orbit_setter
    def set_from_state(self, state, orbit):
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            h_vector = orbit.h * np.array([np.sin(orbit.ascending_node) * np.sin(orbit.inclination),
                                           -np.cos(orbit.ascending_node) * np.sin(orbit.inclination),
                                           np.cos(orbit.inclination)
                                           ])
            self.value = frames.Vector(orbit, state, h_vector, frames.InertialFrame)
        elif self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[1]):
            h = state.position.inertial().cross(state.velocity.inertial())
            self.value = frames.Vector(orbit, state, h, frames.InertialFrame)


class ArgumentOfPeriapsis(OrbitValue):
    symbol = 'arg_periapsis'

    def __init__(self):
        super().__init__(units.rad)
        self.orbit_requirements = [
            ('e'),
            ('e_vec', 'e'),
            ('angular_momentum', 'ascending_node_vec')

        ]
        self.orbit_state_requirements = [
            ('arg_latitude', 'ta'),
        ]

    @orbit_setter
    def set(self, orbit):
        if self.satisfied(orbit, self.orbit_requirements[0]) and orbit.circular():
            self.value = 0

        elif self.satisfied(orbit, self.orbit_requirements[1]) and orbit.equitorial():
            self.value = np.arctan2(orbit.e_vec[1], orbit.e_vec[0])

        elif self.satisfied(orbit, self.orbit_requirements[2]) and not orbit.equitorial() and not orbit.circular():
            e = orbit.e_vec.inertial().value.m
            h_xyz = orbit.angular_momentum.inertial().unit()
            line_of_nodes = orbit.ascending_node_vec
            self.value = np.arctan2(e.dot(np.cross(h_xyz, line_of_nodes)), e.dot(line_of_nodes))

    @orbit_setter
    def set_from_state(self, state, orbit):
        # r(1+ecos(ta))
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            self.value = state.arg_latitude - state.ta


class LongitudeOfAscendingNode(OrbitValue):
    symbol = 'ascending_node'

    def __init__(self):
        super().__init__(units.rad)
        self.orbit_requirements = [
            ('inclination'),
            ('ascending_node_vec')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):
        # special case if i=0
        if self.satisfied(orbit, self.orbit_requirements[0]) and orbit.equitorial():
            self.value = 0

        # Angle from line of nodes
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = np.arctan2(orbit.ascending_node_vec[1], orbit.ascending_node_vec[0])


class AscendingNodeVector(OrbitValue):
    symbol = 'ascending_node_vec'

    def __init__(self):
        super().__init__(units.dimensionless)
        self.orbit_requirements = [
        ]
        self.orbit_state_requirements = [
            ('angular_momentum')
        ]

    @orbit_setter
    def set(self, orbit):
        return False

    @orbit_setter
    def set_from_state(self, state, orbit):
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            h_xyz = orbit.angular_momentum.inertial().unit()
            value = np.cross([0, 0, 1], h_xyz)
            self.value = frames.Vector(orbit, state, value, frames.InertialFrame)


class Inclination(OrbitValue):
    symbol = 'inclination'

    def __init__(self):
        super().__init__(units.rad)
        self.orbit_requirements = [
            ('angular_momentum')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):
        if self.satisfied(orbit, self.orbit_requirements[0]):
            h_xyz = orbit.angular_momentum.unit()
            self.value = np.arccos(h_xyz[2])


class SemimajorAxis(OrbitValue):
    symbol = 'a'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('ra', 'rp'),
            ('se'),
            ('p', 'e')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):

        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = 0.5 * (orbit.ra + orbit.rp)

        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = -orbit.central_body.mu / (2. * orbit.se)

        elif self.satisfied(orbit, self.orbit_requirements[2]):
            self.value = orbit.p / (1. - orbit.e ** 2)

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False


class SemiminorAxis(OrbitValue):
    symbol = 'b'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('p', 'e'),
            ('a', 'e')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):

        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = orbit.p / (1. + orbit.e ** 2.)

        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = orbit.a * np.sqrt(1 - orbit.e ** 2)

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False


class OrbitalPeriod(OrbitValue):
    symbol = 'period'

    def __init__(self):
        super().__init__(units.seconds)
        self.orbit_requirements = [
            ('n')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = 2.0 * np.pi / orbit.n

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False


class MeanMotion(OrbitValue):
    symbol = 'n'

    def __init__(self):
        super().__init__(units.rad / units.second)
        self.orbit_requirements = [
            ('a')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = np.sqrt(orbit.central_body.mu / (orbit.a ** 3))

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False


class SpecificEnergy(OrbitValue):
    symbol = 'se'

    def __init__(self):
        super().__init__(units.km ** 2 / units.s ** 2)
        self.orbit_requirements = [
            ('a')
        ]
        self.orbit_state_requirements = [
            ('r', 'v')
        ]

    @orbit_setter
    def set(self, orbit):

        # -mu/(2a)
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = -orbit.central_body.mu / (2 * orbit.a)

    @orbit_setter
    def set_from_state(self, state, orbit):

        # v^2/2 - mu/r
        if self.state_orbit_satisfied(state, orbit, self.orbit_state_requirements[0]):
            self.value = state.v ** 2. / 2. - orbit.central_body.mu / state.r


class Apoapsis(OrbitValue):
    symbol = 'ra'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('p', 'e'),
            ('a', 'e')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):

        # p / (1 - e)
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = orbit.p / (1 - orbit.e)

        # a * (1 + e)
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = orbit.a * (1 + orbit.e)

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False


class Periapsis(OrbitValue):
    symbol = 'rp'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('p', 'e'),
            ('a', 'e')
        ]
        self.orbit_state_requirements = [
        ]

    @orbit_setter
    def set(self, orbit):

        # p / (1 + e)
        if self.satisfied(orbit, self.orbit_requirements[0]):
            self.value = orbit.p / (1 + orbit.e)

        # a * (1 - e)
        elif self.satisfied(orbit, self.orbit_requirements[1]):
            self.value = orbit.a * (1 - orbit.e)

    @orbit_setter
    def set_from_state(self, state, orbit):
        return False



########### Standard ###########
import logging

########### Local ###########
from orpytal import units, Orbit, KeplarianState, bodies, plotting

########### External ###########
import matplotlib.pyplot as plt
import numpy as np

np.set_printoptions(precision=4)

orbit = Orbit(bodies.earth, name='Orbit1')
state = KeplarianState(orbit, name='State1')

orbit2 = Orbit(bodies.earth, name='Orbit1 Check')
state2 = KeplarianState(orbit2, name='State1 Check')

logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)

if __name__ == '__main__':
    # state.position = frames.Vector(orbit2, state, np.array([106780, -152498, 0])*units.km, frames.InertialFrame)
    # state.velocity = frames.Vector(orbit2, state, np.array([0.4102, 0.5637, 0])*units('km/s'), frames.InertialFrame)

    orbit.e = 0.8
    orbit.p = 37800 * units.km
    orbit.inclination = 45 * units.deg
    orbit.ascending_node = 75 * units.deg
    orbit.arg_periapsis = 30 * units.deg
    # state.ta = -175 * units.deg

    # state.r = 37800 * units.km
    print('------------ Orbit 1 ------------')
    print(orbit)

    print('------------ State 1 ------------')
    print(state)

    orbit2.e = orbit.e
    orbit2.p = orbit.p
    orbit2.inclination = orbit.inclination
    orbit2.arg_periapsis = orbit.arg_periapsis
    orbit2.ascending_node = 120 * units.deg
    plt.figure(1)
    frame = 'inertial'
    plotting.plot_orbit(orbit, frame=frame, planar=False)
    plotting.plot_orbit(orbit2, frame=frame, planar=False)
    plotting.plot_primary(orbit)
    # plotting.plot_state(state, frame=frame)
    plt.show(block=False)
    #
    # state.position.inertial()
    # plt.figure(2)
    # plotting.animate_orbit_fixed(orbit)

    pass



########### Standard ###########
import unittest
import logging

########### Local ###########
from orpytal import Orbit, KeplarianState
from orpytal.common import units
from orpytal.planet_constants import earth

########### External ###########
import numpy as np

logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)

class TestOrbit(unittest.TestCase):
    def test_ta(self):
        orbit = Orbit(earth)
        state = KeplarianState(orbit)

        orbit.e = 0.1
        orbit.p = 18400000 * units.m

        state.r = 18500 * units.km
        state.ascending = True

        print(orbit)
        print(state)

        ta_val = 1.6248767384090719 * units.rad
        assert (np.isclose(state.ta, ta_val))

    def test_from_state(self):
        orbit = Orbit(earth)
        state = KeplarianState(orbit)

        orbit.e = 0.1

        state.r = 18500 * units.km
        state.ta = 90 * units.deg

        orbit.from_state(state)

        assert (np.isclose(orbit.p, 18500.0))
        assert (np.isclose(orbit.h, 85872.6269))


if __name__ == '__main__':
    unittest.main()



########### Standard ###########
import logging

########### Local ###########
from orpytal.base import OrbitBase
from orpytal.common import units, conics_utils, orbit_setter, attribute_setter
from orpytal.errors import ParameterUnavailableError
from orpytal import frames

########### External ###########
import numpy as np


class KeplarianState(object):
    def __init__(self, orbit, name=''):
        self.orbit = orbit
        self.name = name

        self._r = PositionMagnitude()
        self._ta = TrueAnomaly()
        self._arg_latitude = ArgumentOfLatitude()
        self._v = VelocityMagnitude()
        self._position = PositionVector()
        self._velocity = VelocityVector()
        self._fpa = FlightPathAngle()
        self._t_since_rp = TimeSincePeriapsis()
        self._M = MeanAnomaly()
        self._E = EccentricAnomaly()

        self._is_ascending = None

        self.vars = [
            self._r,
            self._ta,
            self._arg_latitude,
            self._v,
            self._position,
            self._velocity,
            self._fpa,
            self._t_since_rp,
            self._M,
            self._E
        ]

    def __str__(self):
        x = ['{} Orbit State {}'.format(self.orbit.name, self.name)]
        for var in self.vars:
            if var.evaluated:
                x.append(str(var))

        return '\n'.join(x)

    def set_vars(self):
        for var in self.vars:
            new_value_set = var.set(self, self.orbit)
            if new_value_set:
                logging.debug('Set {} to {}'.format(var.symbol, var.value))
                self.set_vars()
                break

        orbit_changed = self.orbit.from_state(self)
        if orbit_changed:
            self.set_vars()

    def compare(self, other_state):
        # Compare which vars are evaluated
        evaluated_vars = sorted([v.symbol for v in self.vars if v.evaluated])
        other_evaluated_vars = sorted([v.symbol for v in other_state.vars if v.evaluated])
        for i, var in enumerate(evaluated_vars):
            assert var == other_evaluated_vars[i]

        # Compare values of evaluated variables
        for var in self.vars:
            if var.evaluated and hasattr(other_state, var.symbol):
                if isinstance(var.value, units.Quantity):
                    same = np.isclose(var.value, getattr(other_state, var.symbol))
                elif isinstance(var.value, frames.Vector):

                    vec1 = var.value
                    vec2 = getattr(other_state, var.symbol)
                    same = vec1 == vec2
                else:
                    assert False

            if same:
                logging.debug('Checked {} [✓]'.format(var.symbol))
            else:
                logging.warning('Error Found for {} [x]'.format(var.symbol))
    @property
    def r(self):
        return self._r.value

    @r.setter
    @attribute_setter
    def r(self, r):
        self._r.value = r

    @property
    def ta(self):
        return self._ta.value

    @ta.setter
    @attribute_setter
    def ta(self, ta):
        self._ta.value = ta

    @property
    def arg_latitude(self):
        return self._arg_latitude.value

    @arg_latitude.setter
    @attribute_setter
    def arg_latitude(self, arg_latitude):
        self._arg_latitude.value = arg_latitude

    @property
    def v(self):
        return self._v.value

    @v.setter
    @attribute_setter
    def v(self, v):
        self._v.value = v

    @property
    def position(self):
        return self._position.value

    @position.setter
    @attribute_setter
    def position(self, position):
        self._position.value = position

    @property
    def velocity(self):
        return self._velocity.value

    @velocity.setter
    @attribute_setter
    def velocity(self, velocity):
        self._velocity.value = velocity

    @property
    def fpa(self):
        return self._fpa.value

    @fpa.setter
    @attribute_setter
    def fpa(self, fpa):
        self._fpa.value = fpa

    @property
    def M(self):
        return self._M.value

    @M.setter
    @attribute_setter
    def M(self, M):
        self._M.value = M

    @property
    def E(self):
        return self._E.value

    @E.setter
    @attribute_setter
    def E(self, E):
        self._E.value = E

    @property
    def t_since_rp(self):
        return self._t_since_rp.value

    @t_since_rp.setter
    @attribute_setter
    def t_since_rp(self, t_since_rp):
        self._t_since_rp.value = t_since_rp

    @property
    def ascending_sign(self):
        return 1 if self.is_ascending() else -1

    @property
    def ascending(self):
        return self._is_ascending

    @ascending.setter
    @attribute_setter
    def ascending(self, ascending):
        self._is_ascending = ascending

    def is_ascending(self):

        angle_to_check = None

        if self._ta.evaluated:
            angle_to_check = self.ta
        elif self._fpa.evaluated:
            angle_to_check = self.fpa
        elif self._E.evaluated:
            angle_to_check = self.E
        elif self._velocity.evaluated:
            self._is_ascending = self.velocity.rotating().value[0].m > 0
        elif self.ascending is None:
            raise ParameterUnavailableError("Need either ta, fpa or E to check ascending")

        if angle_to_check is not None:
            self._is_ascending = 0 <= angle_to_check <= np.pi

        return self.ascending

    def angle_check_tan(self, tan_val):
        if (self.is_ascending() and tan_val >= 0) or (not self.is_ascending() and tan_val < 0):
            return tan_val
        else:
            return np.pi + tan_val


class StateValue(OrbitBase):

    def set(self, state, orbit):
        raise NotImplementedError()

    def satisfied(self, state, orbit, requirements):
        return self.state_orbit_satisfied(state, orbit, requirements)


class TrueAnomaly(StateValue):
    symbol = 'ta'

    def __init__(self):
        super().__init__(units.radian)
        self.orbit_requirements = [
            ('e', 'p', 'r', 'fpa'),
            ('e', 'E'),
            ('r', 'v', 'fpa'),
            ('angular_momentum', 'position', 'e_vec', 'e', 'inclination'),
            ('angular_momentum', 'position', 'ascending_node_vec', 'e', 'inclination'),
            ('position', 'e', 'inclination'),
            ('angular_momentum', 'position', 'e_vec', 'e', 'inclination', 'ascending_node_vec'),
        ]

    @orbit_setter
    def set(self, state, orbit):

        # acos((1/e) (p/r - 1))
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            cos_val = (1. / orbit.e) * (orbit.p / state.r - 1.)
            self.value = state.ascending_sign * np.arccos(cos_val)

        # acos(2 atan(sqrt( (1+e)/(1-e) tan(E/2) )))
        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            ta = 2 * np.arctan(np.sqrt((1 + orbit.e) / (1 - orbit.e)) * np.tan(state.E / 2))
            self.value = state.angle_check_tan(ta)

        # atan( ((rv2)/mu cos(g)sin(g)) / ((rv2)/mu cos^2(g)-1) )
        elif self.satisfied(state, orbit, self.orbit_requirements[2]):
            t = (state.r * state.v ** 2) / orbit.central_body.mu
            ta = np.arctan((t * np.cos(state.fpa) * np.sin(state.fpa)) /
                           (t * np.cos(state.fpa) ** 2 - 1))
            self.value = state.angle_check_tan(ta)

        elif self.satisfied(state, orbit, self.orbit_requirements[3]) and orbit.equitorial() and not orbit.circular():

            h = orbit.angular_momentum.inertial().unit()
            e = orbit.e_vec
            r = state.position.inertial()
            self.value = np.arctan2(h.dot(e.cross(r)), r.dot(e))

        elif self.satisfied(state, orbit, self.orbit_requirements[4]) and not orbit.equitorial() and orbit.circular():
            h = orbit.angular_momentum.inertial().unit()
            r = state.position.inertial()
            n = orbit.ascending_node_vec.inertial()
            self.value = np.arctan2(r.dot(np.cross(h, n)), r.dot(n))

        elif self.satisfied(state, orbit, self.orbit_requirements[5]) and orbit.equitorial() and orbit.circular():
            r = state.position.inertial()
            self.value = np.arctan2(r[1], r[0])

        elif self.satisfied(state, orbit, self.orbit_requirements[6]) and not orbit.equitorial() and not orbit.circular():

            # unit vectors
            ehat = orbit.e_vec.inertial().unit()
            hhat = orbit.angular_momentum.inertial().unit()
            phat = np.cross(hhat, ehat)

            r = state.position.inertial().unit()
            y = r.dot(phat)
            x = r.dot(ehat)

            self.value = np.arctan2(y, x)

class PositionMagnitude(StateValue):
    symbol = 'r'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('p', 'e', 'ta'),
            ('a', 'e', 'E'),
            ('a', 'E', 'H'),
            ('position')
        ]

    @orbit_setter
    def set(self, state, orbit):

        # p/1+ecos(ta)
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = orbit.p / (1.0 + orbit.e * np.cos(state.ta))

        # a (1-ecosE)
        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            self.value = orbit.a * (1 - orbit.e * np.cos(state.E))

        # |a|(E cosh H - 1)
        elif self.satisfied(state, orbit, self.orbit_requirements[2]):
            self.value = abs(orbit.a) * (state.E * np.cosh(state.H) - 1)

        # |r|
        elif self.satisfied(state, orbit, self.orbit_requirements[3]):
            self.value = np.linalg.norm(state.position.value)


class ArgumentOfLatitude(StateValue):
    symbol = 'arg_latitude'

    def __init__(self):
        super().__init__(units.rad)
        self.orbit_requirements = [
            ('arg_periapsis', 'ta'),
            ('inclination', 'position', 'angular_momentum')
        ]

    @orbit_setter
    def set(self, state, orbit):
        # omega + ta
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = orbit.arg_periapsis + state.ta

        # from DCM
        if self.satisfied(state, orbit, self.orbit_requirements[1]):
            rhat = state.position.inertial().unit()
            hhat = orbit.angular_momentum.inertial().unit()
            theta_hat = np.cross(hhat, rhat)

            # special case i=0
            if orbit.inclination == 0 * units.rad:
                self.value = 0
            else:
                sin_val = np.arcsin(rhat[2] / np.sin(orbit.inclination.to(units.rad)))
                cos_val = np.arccos(theta_hat[2] / np.sin(orbit.inclination.to(units.rad)))
                self.value = conics_utils.common_val(sin_val, cos_val)


class FlightPathAngle(StateValue):
    symbol = 'fpa'

    def __init__(self):
        super().__init__(units.rad)
        self.orbit_requirements = [
            ('h', 'r', 'v'),
            ('velocity')
        ]

    @orbit_setter
    def set(self, state, orbit):

        # acos(h/(rv))
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = state.ascending_sign * np.arccos(orbit.h / (state.r * state.v))

        # atan(vr/vt)
        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            v = state.velocity.orbit_fixed()
            fpa = np.arctan(v[0] / v[1])
            self.value = state.angle_check_tan(fpa)

        if self.value and np.isnan(self.value):
            self.value = 0

class VelocityMagnitude(StateValue):
    symbol = 'v'

    def __init__(self):
        super().__init__(units.km / units.second)
        self.orbit_requirements = [
            ('r', 'se'),
            ('r', 'a'),
            ('velocity',)
        ]

    @orbit_setter
    def set(self, state, orbit):

        # sqrt( 2(se + mu/r))
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = np.sqrt(2 * (orbit.se + orbit.central_body.mu / state.r))
        # sqrt( mu(2/r-1/a) )
        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            self.value = np.sqrt(orbit.central_body.mu * (2. / state.r - 1 / orbit.a))
        # |v|
        elif self.satisfied(state, orbit, self.orbit_requirements[2]):
            self.value = state.velocity.norm()


class PositionVector(StateValue):
    symbol = 'position'

    def __init__(self):
        super().__init__(units.km)
        self.orbit_requirements = [
            ('r')
        ]

    @orbit_setter
    def set(self, state, orbit):
        # r (rhat)
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = frames.Vector(orbit,
                                       state,
                                       np.array([state.r.m, 0, 0]),
                                       frames.RotatingFrame)


class VelocityVector(StateValue):
    symbol = 'velocity'

    def __init__(self):
        super().__init__(units.km / units.second)
        self.orbit_requirements = [
            ('ta', 'e', 'h'),
            ('fpa' 'v')
        ]

    @orbit_setter
    def set(self, state, orbit):

        # Rotating Frame
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            vr = (orbit.central_body.mu * orbit.e) / orbit.h * np.sin(state.ta)
            vt = orbit.central_body.mu / orbit.h * (1 + orbit.e * np.cos(state.ta))
            self.value = frames.Vector(orbit,
                                       state,
                                       np.array([vr.m, vt.m, 0]) * vr.u,
                                       frames.RotatingFrame)

        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            self.value = frames.Vector(orbit,
                                       state,
                                       np.array([state.v * np.sin(state.fpa), state.v * np.cos(state.fpa), 0]),
                                       frames.RotatingFrame)


class TimeSincePeriapsis(StateValue):
    symbol = 't_since_rp'

    def __init__(self):
        super().__init__(units.seconds)
        self.orbit_requirements = [
            ('E', 'e', 'n'),
        ]

    @orbit_setter
    def set(self, state, orbit):
        # Rotating Frame
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = (state.E - orbit.e * np.sin(state.E)) / orbit.n


class MeanAnomaly(StateValue):
    symbol = 'M'

    def __init__(self):
        super().__init__(units.radians)
        self.orbit_requirements = [
            ('n', 't_since_rp'),
            ('E', 'e')
        ]

    @orbit_setter
    def set(self, state, orbit):
        # n * ttp
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            self.value = orbit.n * state.t_since_rp

        # E - e sin(E)
        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            self.value = state.E - orbit.e * np.sin(state.E)


class EccentricAnomaly(StateValue):
    symbol = 'E'

    def __init__(self):
        super().__init__(units.radians)
        self.orbit_requirements = [
            ('a', 'r', 'e'),
            ('e', 'M')
        ]

    @orbit_setter
    def set(self, state, orbit):
        # acos((a-r)/(ae))
        if self.satisfied(state, orbit, self.orbit_requirements[0]):
            cos_val = (orbit.a - state.r) / (orbit.a * orbit.e)
            if cos_val > 1 and np.isclose(cos_val, 1):
                self.value = 0
            else:
                self.value = state.ascending_sign * np.arccos(
                    (orbit.a - state.r) / (orbit.a * orbit.e))

        elif self.satisfied(state, orbit, self.orbit_requirements[1]):
            self.value = self._iterative_eccentric_anomaly(state, orbit)

    def _iterative_eccentric_anomaly(self, state, orbit, **kwargs):
        return self._find_eccentric_anomaly_recursively(orbit, state.M, state.M, **kwargs)

    def _find_eccentric_anomaly_recursively(self, orbit, E, M, tol=1e-12):
        E1 = E - (E - orbit.e * np.sin(E) - M) / (1 - orbit.e * np.cos(E))
        dE = np.fabs(E - E1)

        if dE < tol:
            return E

        return self._find_eccentric_anomaly_recursively(orbit, E1, M, tol=tol)



import numpy as np
import scipy as sp
import math
import os
import pickle


def newton_raphson(fx, dfx, ic, rtol, max_iterations, extra_params=(), attenuation=1.0, atol=1e-6):
    val_list = [ic]

    val = val_list[0]

    for n in range(max_iterations):
        # Find f(x) value
        fx_val = fx(val, *extra_params)

        # Find f'(x) value
        dfdx_val = dfx(val, *extra_params)

        # Do not continue if f'(x)=0
        if not dfdx_val:
            raise ValueError('Cannot continue for f\'(x)=0')

        # Perform Newton-Raphson step
        new_val = val - attenuation * (fx_val / dfdx_val)

        # Add to iterative list (in case we want to see entire list at some point)
        val_list.append(new_val)

        # Stop if we are within the tolerance, else continue iterations
        if np.abs(new_val - val) < rtol and np.abs(fx_val) < atol:
            # print 'Solution found in %i iterations' % len(val_list)
            return val_list[-1]
        else:
            val = new_val

    # Return None if nothing is found
    raise ValueError('No solution found in %i iterations' % max_iterations)


def pickle_dict(filename, pdict):
    file_object = open(filename, 'wb')
    pickle.dump(pdict, file_object)
    file_object.close()


def load_pickle(filename):
    if os.path.getsize(filename) == 0:
        print('{} is an empty file!'.format(filename))
    filename_bytes = filename.encode()
    try:
        file_object = open(filename_bytes, 'rb')
        return pickle.load(file_object)
    except UnicodeDecodeError as ude:
        file_object = open(filename_bytes, 'rb')
        return pickle.load(file_object, encoding='latin1')



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

print('-----------------------------------------')
print(find_packages(exclude=['tests']))
print('-----------------------------------------')

# https://blog.ionelmc.ro/2014/05/25/python-packaging/
setup(
    name="orpytal",
    version="0.0.11",
    description="Python Package for Two-Body Orbital Mechanics",
    author="Nick LaFarge",
    author_email="nick.lafarge@gmail.com",
    url="http://nicklafarge.space/orpytal",
    download_url="https://github.com/nicklafarge/OrPytal",
    license="MIT",
    keywords=[
        "aero", "aerospace", "engineering",
        "astrodynamics", "orbits", "kepler", "orbital mechanics"
    ],
    python_requires=">=3.5",
    install_requires=[
        "numpy",
        "matplotlib>=2.0",
        "seaborn",
        "matlab",
        "ipython",
        "untangle",
        "pint"
    ],
    # packages=find_packages(exclude=['tests']),
    packages=['orpytal'],
    package_data={'orpytal': ['data/body_data.xml']},
    include_package_data=True,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    long_description=open('README.md').read(),
    zip_safe=False,
)
